import os
import sys
import getpass
import logging
import configparser
import argparse
import pprint
import re
import getpass
import tempfile
import shutil
import pycurl

from io import BytesIO
from munch import Munch
from cached_property import cached_property

from .utils import run, QUIET, MODEST, VERBOSE,\
        parse_config_rpmdefines, extract_srpm, is_upload_item, is_reserved_item,\
        DoNothing, has_scheme, download_file
from .rpm_package import RpmPackage
from .lookaside_cache import LookasideCache
from .git_client import GitClient
from .errors import RpkgError

log = logging.getLogger('__main__')
log.setLevel(logging.INFO)
log.addHandler(logging.StreamHandler(sys.stdout))

class RpkgUtil(object):

    def __init__(self, name='rpkg'):
        self.name = name
        self.glob_conf = '/etc/{0}.conf'.format(name)
        self.user_conf = os.path.expanduser('~/.config/{0}.conf'.format(name))
        self.path = os.path.abspath(self.args.path)
        try:
            if os.path.isdir(self.args.spec):
                self.pkg_path = self.args.spec
            else:
                self.pkg_path = os.path.dirname(self.args.spec) or os.getcwd()
        except:
            self.pkg_path = self.path

    def main(self):
        self.setup_log()

        try:
            os.chdir(self.path)
            cmd = getattr(self.args, 'command', self.parser.print_help)
            cmd()
        except Exception as e:
            if self.args.verbosity == VERBOSE:
                raise
            else:
                err_msg = str(e)
                if err_msg:
                    log.error(err_msg)
                sys.exit(1)
        except KeyboardInterrupt:
            pass

    ######################## setup ########################

    def setup_log(self):
        if self.args.verbosity == VERBOSE:
            log.setLevel(logging.DEBUG)
        elif self.args.verbosity == QUIET:
            log.setLevel(logging.WARNING)
        else:
            log.setLevel(logging.INFO)

    @cached_property
    def args(self):
        return self.parser.parse_args()

    @cached_property
    def git_props(self):
        return GitClient.git_props(self.pkg_path)

    @cached_property
    def cfg_parser(self):
        parser = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation())
        default_conf = os.path.join(os.path.dirname(__file__), 'default.conf')

        if self.args.config_file:
            configs_to_read = [default_conf, self.args.config_file]
        else:
            configs_to_read = [default_conf, self.glob_conf, self.user_conf]

            if self.git_props.root:
                repo_conf = os.path.join(self.git_props.root, 'rpkg.conf')
                configs_to_read.append(repo_conf)

            pkg_conf = os.path.join(self.pkg_path, 'rpkg.conf')
            configs_to_read.append(pkg_conf)

        parser.read(configs_to_read)
        parser['DEFAULT']['user'] = self.args.user
        parser['DEFAULT']['path'] = self.path
        parser['DEFAULT']['pkg_path'] = self.pkg_path

        parser.add_section('git_props')
        for git_prop, value in self.git_props.items():
            parser['git_props'][git_prop] = value

        for key_val in self.args.config:
            key, val = re.split(r'\s*=\s*', key_val, maxsplit=1)
            if '.' in key:
                section, param = re.split(r'\.', key, maxsplit=1)
            else:
                section, param = 'DEFAULT', key

            if param not in parser[section]:
                raise RpkgError("Invalid key {} passed.".format(key))

            parser[section][param] = val

        return parser

    @cached_property
    def cfg(self):
        return Munch({
            'DEFAULT': Munch({
                'user': self.cfg_parser.get('DEFAULT', 'user'),
                'path': os.path.expanduser(self.cfg_parser.get('DEFAULT', 'path')),
                'pkg_path': os.path.expanduser(self.cfg_parser.get('DEFAULT', 'pkg_path')),
            }),
            'git_props': Munch({
                'root': self.cfg_parser.get('git_props', 'root'),
                'branch': self.cfg_parser.get('git_props', 'branch'),
                'remote': self.cfg_parser.get('git_props', 'remote'),
                'remote_url': self.cfg_parser.get('git_props', 'remote_url'),
                'remote_netloc': self.cfg_parser.get('git_props', 'remote_netloc'),
                'head': self.cfg_parser.get('git_props', 'head'),
                'head_short': self.cfg_parser.get('git_props', 'head_short'),
                'status': self.cfg_parser.get('git_props', 'status'),
                'merged_tag_refs': self.cfg_parser.get('git_props', 'merged_tag_refs'),
                'submodule_refs': self.cfg_parser.get('git_props', 'submodule_refs'),
                'dirty': self.cfg_parser.get('git_props', 'dirty'),
            }),
            'rpkg': Munch({
                'preprocess_spec': self.cfg_parser.getboolean('rpkg', 'preprocess_spec'),
                'base_output_path': os.path.expanduser(self.cfg_parser.get('rpkg', 'base_output_path')),
                'user_macros': utils.parse_config_list(self.cfg_parser.get('rpkg', 'user_macros')),
                'bcond_with': utils.parse_config_list(self.cfg_parser.get('rpkg', 'with')),
                'bcond_without': utils.parse_config_list(self.cfg_parser.get('rpkg', 'without')),
                'rpmdefines': parse_config_rpmdefines(self.cfg_parser.get('rpkg', 'rpmdefines')),
                'bump_rpmdefines': parse_config_rpmdefines(self.cfg_parser.get('rpkg', 'bump_rpmdefines')),
                'rpmlint_config': os.path.expanduser(self.cfg_parser.get('rpkg', 'rpmlint_config')),
                'default_macros': utils.parse_config_list(self.cfg_parser.get('rpkg', 'default_macros')),
                'copr_project': self.cfg_parser.get('rpkg', 'copr_project'),
            }),
            'git': Munch({
                'clone_url': self.cfg_parser.get('git', 'clone_url'),
                'anon_clone_url': self.cfg_parser.get('git', 'anon_clone_url'),
                'default_namespace': self.cfg_parser.get('git', 'default_namespace'),
                'push_follow_tags': self.cfg_parser.get('git', 'push_follow_tags'),
                'clean_force': self.cfg_parser.get('git', 'clean_force'),
                'clean_dirs': self.cfg_parser.get('git', 'clean_dirs'),
                'log_graph': self.cfg_parser.get('git', 'log_graph'),
                'log_all': self.cfg_parser.get('git', 'log_all'),
                'gpg_keyid': self.cfg_parser.get('git', 'gpg_keyid'),
            }),
            'lookaside': Munch({
                'download_url': self.cfg_parser.get('lookaside', 'download_url'),
                'upload_url': self.cfg_parser.get('lookaside', 'upload_url'),
                'client_cert': self.cfg_parser.get('lookaside', 'client_cert'),
                'client_key': self.cfg_parser.get('lookaside', 'client_key'),
                'ca_cert': self.cfg_parser.get('lookaside', 'ca_cert'),
            }),
        })

    def read_spec_file(self, spec_path):
        return open(spec_path, 'r').read()

    @cached_property
    def rpm_package(self):
        spec_path = getattr(self.args, 'spec', None)

        if not spec_path or not os.path.isfile(spec_path):
            spec_path = RpmPackage.locate_spec(self.pkg_path)

        spec_path = os.path.abspath(spec_path)

        if self.git_props.head and not self.git_props.dirty:
            try:
                spec_content = self.git_client.read_file(
                    self.git_props.head, spec_path)
            except:
                spec_content = self.read_spec_file(spec_path)
        else:
            spec_content = self.read_spec_file(spec_path)

        return RpmPackage(self.pkg_path, spec_content, spec_path,
                          getattr(self.args, 'outdir', None),
                          **self.cfg.get('rpkg'),
                          git_props=self.git_props,
                          verbosity=self.args.verbosity)

    @cached_property
    def git_client(self):
        return GitClient(self.pkg_path,
                         **self.cfg.get('git'),
                         git_props=self.git_props,
                         verbosity=self.args.verbosity)

    @cached_property
    def lookaside(self):
        return LookasideCache(self.pkg_path,
                              **self.cfg.get('lookaside'),
                              verbosity=self.args.verbosity)

    @cached_property
    def parser(self):
        """Setup the argument parser and register some basic commands."""
        parser = argparse.ArgumentParser(
            prog='rpkg', epilog='For detailed help pass --help to a target subcommand.')

        parser.add_argument('--config', '-c', metavar='KEY=VAL', default=[],
                            action='append', help='Set specific config '
                            'parameter KEY to a value VAL. May be specified '
                            'multiple times. Use <section>.<param> format '
                            'for KEY. Takes effect after -f option.')
        parser.add_argument('--config-file', '-C', action='store', default=None,
                            help='Specify a unique config file to be used.')
        parser.add_argument('--user', '-u', default=getpass.getuser(),
                            help='Override the discovered system user name.')
        parser.add_argument('--path', '-p', action='store', default=os.getcwd(),
                            help='Define the directory to work in (defaults to cwd). '
                            'The working directory primarily determines where spec '
                            'file, sources file, and rpkg.conf are located. These '
                            'files are what forms an rpm (sub)package. The path '
                            'to a subpackage is additionally changed by --spec '
                            'argument, if specified.')
        parser.add_argument('--verbose', '-v', dest='verbosity',
                            action='store_const', const=VERBOSE, default=MODEST,
                            help='Run with verbose debug output.')
        parser.add_argument('--quiet', '-q', dest="verbosity",
                            action='store_const', const=QUIET, default=MODEST,
                            help='Run quietly only displaying errors.')

        self.subparsers = self.setup_subparsers(parser)
        return parser

    def setup_subparsers(self, parser):
        sp = parser.add_subparsers(
            help='Available subcommands')

        self.register_add(sp)
        self.register_clean(sp)
        self.register_clog(sp)
        self.register_clone(sp)
        self.register_config(sp)
        self.register_commit(sp)
        self.register_compile(sp)
        self.register_copr_build(sp)
        self.register_diff(sp)
        self.register_gimmespec(sp)
        self.register_import_srpm(sp)
        self.register_install(sp)
        self.register_lint(sp)
        self.register_local(sp)
        self.register_log(sp)
        self.register_merge(sp)
        self.register_new(sp)
        self.register_nvr(sp)
        self.register_prep(sp)
        self.register_pull(sp)
        self.register_push(sp)
        self.register_reset(sp)
        self.register_sources(sp)
        self.register_spec(sp)
        self.register_srpm(sp)
        self.register_status(sp)
        self.register_switch(sp)
        self.register_tag(sp)
        self.register_unused_patches(sp)
        self.register_upload(sp)
        self.register_verify_files(sp)

        return sp

    ######################## commands ########################

    def add(self):
        self.git_client.add(self.args.all,
                            self.args.paths)

    def config(self):
        self.cfg_parser.write(sys.stdout)

    def clean(self):
        self.git_client.clean(self.args.dry_run,
                              self.args.x,
                              self.args.force,
                              self.args.d)

    def clog(self):
        print(self.rpm_package.clog(self.args.raw))

    def clone(self):
        if self.args.with_dirs:
            self.git_client.clone_with_dirs(self.args.repo_path,
                                            self.args.directory,
                                            self.args.anonymous)
        else:
            self.git_client.clone(self.args.repo_path,
                                  self.args.directory,
                                  self.args.branch,
                                  self.args.anonymous)

    def commit(self):
        msg_lines = []
        if self.args.message:
            msg_lines += [self.args.message]

            if self.args.with_changelog:
                msg_lines += self.rpm_package.clog(raw=True).split('\n')

        self.git_client.commit(self.args.all, self.args.amend, msg_lines,
                               self.args.file, self.args.signoff, self.args.files)

        if self.args.tag:
            self.rpm_package.version_bump = True
            self.rpm_package.release_bump = True
            tagname = self.rpm_package.nvrid()
            self.git_client.add_tag(tagname)

        if self.args.push:
            self.git_client.push()

    def compile(self):
        self.lookaside.sources(self.git_client.repo_path(),
                               raise_on_missing_sources=False)

        self.rpm_package.compile(target=self.args.target,
                         short=self.args.short_circuit,
                         nocheck=self.args.nocheck,
                         bcond_with=getattr(self.args, 'with'),
                         bcond_without=self.args.without)

    def copr_build(self):
        self.rpm_package.copr_build(self.args.project,
                                    not self.args.wait,
                                    self.args.copr_config)

    def diff(self):
        self.git_client.diff(self.args.cached, self.args.files)

    def gimmespec(self):
        print(self.rpm_package.spec_path)

    def import_srpm(self):
        if self.git_props.dirty:
            raise RpkgError('There are uncommitted changes in your repo.')

        tempdir = tempfile.mkdtemp(
            dir=self.cfg.rpkg.base_output_path, prefix='tmpdir_')

        if has_scheme(self.args.srpm):
            tmp_srpm = tempfile.NamedTemporaryFile(
                'w', prefix='tmp_', dir=self.cfg.rpkg.base_output_path, delete=False)
            tmp_srpm.close()

            log.info('Downloading: ' + self.args.srpm)
            status = download_file(self.args.srpm, tmp_srpm.name, max_redirs=3)

            if status != 200:
                shutil.rmtree(tempdir)
                os.remove(tmp_srpm.name)
                raise RpkgError('Server returned status code %d.' % status)

            extract_srpm(tmp_srpm.name, tempdir)
            os.remove(tmp_srpm.name)
        else:
            extract_srpm(self.args.srpm, tempdir)

        if self.git_props.root:
            current_items = self.git_client.ls_tree(
                self.git_props.head, '--name-only').split()
            current_items += [entry.filename for entry
                              in self.lookaside.read_source_entries()]
        else:
            current_items = os.listdir(self.pkg_path)

        new_items = os.listdir(tempdir)

        removed_nonupload_items = []
        for item in set(current_items) - set(new_items):
            item_path = os.path.join(self.pkg_path, item)
            if not os.path.isfile(item_path) or is_reserved_item(item, self.name):
                continue

            log.info('Removing {0}.'.format(item))
            if not is_upload_item(item):
                removed_nonupload_items.append(item)
            os.remove(item_path)

        upload_items = []
        for item in new_items:
            if is_upload_item(item):
                upload_items.append(item)

            source = os.path.join(tempdir, item)
            if not os.path.isfile(source):
                raise RpkgError('Only files expected in extracted srpm content.')

            dest = os.path.join(self.pkg_path, item)
            log.info('Importing {0}.'.format(item))
            shutil.move(source, dest)

        shutil.rmtree(tempdir)
        repo_path = self.args.repo_path or self.git_client.repo_path()

        source_entries = self.lookaside.upload(
            repo_path, upload_items, offline=self.args.offline)

        self.lookaside.write_source_entries(
            source_entries, append=False)

        if not self.git_props.root:
            return

        self.git_client.write_ignores(
            [entry.filename for entry in source_entries],
            append=False)

        items_to_add = set(['.gitignore', 'sources'])
        items_to_add |= set(removed_nonupload_items)
        items_to_add |= set(new_items)
        items_to_add -= set(upload_items)
        self.git_client.add_to_index(*items_to_add)

        if not self.args.skip_diffs:
            self.git_client.show_diff_index()

        log.info('Please, don\'t forget to commit.')

    def install(self):
        self.lookaside.sources(self.git_client.repo_path(),
                               raise_on_missing_sources=False)

        self.rpm_package.install(target=self.args.target,
                                 short=self.args.short_circuit,
                                 nocheck=self.args.nocheck,
                                 bcond_with=getattr(self.args, 'with'),
                                 bcond_without=self.args.without)

    def lint(self):
        self.rpm_package.lint(
            self.args.info,
            self.args.rpmlint_config,
            self.args.target_path)

    def local(self):
        self.lookaside.sources(self.git_client.repo_path(),
                               raise_on_missing_sources=False)

        self.rpm_package.local(target=self.args.target,
                               nocheck=self.args.nocheck,
                               bcond_with=getattr(self.args, 'with'),
                               bcond_without=self.args.without)

    def log(self):
        self.git_client.log(self.args.graph,
                            self.args.all,
                            self.args.show_signature,
                            self.args.positionals)

    def merge(self):
        if self.args.cont:
            self.git_client.merge_continue()
        elif self.args.abort:
            self.git_client.merge_abort()
        else:
            self.git_client.merge(self.args.no_commit, self.args.squash,
                                  self.args.message, self.args.file,
                                  self.args.signoff, self.args.commits)

    def new(self):
        package_name = self.rpm_package.nvre().name
        self.git_client.new(package_name)

    def nvr(self):
        print(self.rpm_package.nvrid())

    def prep(self):
        self.lookaside.sources(self.git_client.repo_path(),
                               raise_on_missing_sources=False)

        self.rpm_package.prep(target=self.args.target,
                      bcond_with=getattr(self.args, 'with'),
                      bcond_without=self.args.without)

    def pull(self):
        self.git_client.pull(self.args.rebase,
                             self.args.no_rebase)

    def push(self):
        self.git_client.push(self.args.force, self.args.follow_tags)

    def reset(self):
        if self.args.mixed:
            mode = 1
        elif self.args.soft:
            mode = 2
        elif self.args.hard:
            mode = 3
        else:
            mode = 0

        self.git_client.reset(mode, self.args.positional)

    def sources(self):
        repo_path = self.args.repo_path or self.git_client.repo_path()
        self.lookaside.sources(repo_path, self.args.outdir)

    def spec(self):
        self.rpm_package.produce_sources = self.args.sources

        if self.args.print:
            final_spec_text = self.rpm_package.final_spec_text()
            print(final_spec_text)
        else:
            self.rpm_package.final_spec()

    def srpm(self):
        self.lookaside.sources(self.git_client.repo_path(),
                               raise_on_missing_sources=False)

        self.rpm_package.srpm(target=getattr(self.args, 'target', None),
                              bcond_with=getattr(self.args, 'with', []),
                              bcond_without=getattr(self.args, 'without', []))

    def status(self):
        self.git_client.status(self.args.pathspecs)

    def switch(self):
        if self.args.list_branches:
            self.git_client.list_branches()
            return

        if self.args.fetch:
            self.git_client.fetch()

        if self.args.target:
            self.git_client.switch(self.args.target,
                                   self.args.create,
                                   self.args.detach)
        else:
            self.git_client.list_branches()

    def tag(self):
        self.rpm_package.version_bump = True
        self.rpm_package.release_bump = True

        if self.args.list:
            self.git_client.list_tags(self.args.tag)
            return
        elif self.args.delete:
            if not self.args.tag:
                sys.stderr.write('Please, specify tag to be deleted.\n')
            else:
                self.git_client.delete_tag(self.args.tag)
            return

        tagname = self.args.tag or self.rpm_package.nvrid(
            self.args.name, self.args.version, self.args.release)
        self.git_client.add_tag(tagname, self.args.skip_edit)

    def unused_patches(self):
        unused_patches = self.rpm_package.unused_patches()
        if unused_patches:
            print('\n'.join(unused_patches))

    def upload(self):
        repo_path = self.args.repo_path or self.git_client.repo_path()

        source_entries = self.lookaside.upload(
            repo_path, self.args.files, offline=self.args.offline)
        self.lookaside.write_source_entries(
            source_entries, append=True)

        if not self.git_props.root:
            return

        self.git_client.write_ignores(
            [entry.filename for entry in source_entries],
            append=True)

        self.git_client.add_to_index(
            self.lookaside.sources_filepath,
            self.git_client.gitignore_path)

        if not self.git_client.status(
                self.lookaside.sources_filepath,
                self.git_client.gitignore_path):
            return

        log.info('Please, don\'t forget to commit.')

    def verify_files(self):
        self.rpm_package.verify_files(
            self.args.target_dir)

    ######################## subparsers ########################

    def register_add(self, subparsers):
        """Register the add target"""
        add_parser = subparsers.add_parser(
            'add', description='This command updates the index '
            'with changes found in the working tree.',
            help='Add changes to index')
        add_parser.add_argument(
            '--all', '-A', action='store_true',
            help='Update index with all changes found in the working tree.')
        add_parser.add_argument(
            'paths', action='store', nargs='*', default=[],
            help='Specific paths to add changes for.')
        add_parser.set_defaults(command=self.add)

    def register_config(self, subparsers):
        config_parser = subparsers.add_parser(
            'config', description='Do nothing, only output used config.',
            help='Do nothing, only output used config')
        config_parser.set_defaults(command=self.config)

    def register_clean(self, subparsers):
        """Register the clean target"""
        clean_parser = subparsers.add_parser(
            'clean', description='This command can be used to '
            'clean up your work tree from untracked files and, '
            'files described by .gitignore additionally if -x '
            'option is passed.', help='Remove untracked files')
        clean_parser.add_argument(
            '--dry-run', '-n', action='store_true',
            help='Don\'t actually remove anything, just show '
            'what would be done.')
        clean_parser.add_argument(
            '-x', action='store_true',
            help='Don\'t exclude files described by .gitignore rules '
            'from deletion.')
        clean_parser.add_argument(
            '--force', '-f', action='store_true',
            help='This option is required to perform deletion of '
            'any files or directories if git configuration variable '
            'clean.requireForce is set to true. It is passed '
            'by default if rpkg configuration variable '
            'git.clean_force is set to true.')
        clean_parser.add_argument(
            '-d', action='store_true',
            help='Remove whole directories in addition to files. '
            'This option is passed by default if rpkg configuration '
            'variable git.clean_dirs is set to true.')
        clean_parser.set_defaults(command=self.clean)

    def register_clog(self, subparsers):
        """Register the clog target"""
        clog_parser = subparsers.add_parser(
            'clog', description='This will output the latest '
            'rpm changelog entry to stdout. The leading '
            '"- " text will be stripped.',
            help='Output top changelog entry to stdout')
        clog_parser.add_argument(
            '--raw', action='store_true',
            help='Generate a more "raw" clog without '
            'twiddling the contents')
        clog_parser.set_defaults(command=self.clog)

    def register_clone(self, subparsers):
        """Register the clone target"""
        clone_parser = subparsers.add_parser(
            'clone', description='This command will clone the named '
            'repository from the configured repository base URL. '
            'By default it will also checkout the master branch '
            'for your working copy.',
            help='Clone and checkout a repository')
        clone_parser.add_argument(
            'repo_path', action='store',
            help='Name of the repository to clone '
            'or full clone url. If just a basename is specified, '
            'preconfigured git clone urls and default namespace '
            'is used to construct the final clone url.')
        clone_parser.add_argument(
            'directory', action='store', default=None, nargs='?',
            help='Directory in which to clone the repository')
        clone_parser.add_argument(
            '--branches', '-B', action='store_true', dest='with_dirs',
            help='Checkout a subdirectory for each branch')
        clone_parser.add_argument(
            '--branch', '-b', action='store',
            help='Check out a specific branch. This has no effect '
            'if --branches/-B is specified.')
        clone_parser.add_argument(
            '--anonymous', '-a', action='store_true',
            help='Check out a repository anonymously')
        clone_parser.set_defaults(command=self.clone)

    def register_commit(self, subparsers):
        """Register the commit target and ci alias"""
        commit_parser = subparsers.add_parser(
            'commit', description='This invokes a git commit. '
            'All tracked files with changes will be committed '
            'unless a specific file list is provided. $EDITOR '
            'will be used to generate a changelog message '
            'unless one is given to the command. A tag or '
            'push can be done at the same time. If you have '
            'set git.gpg_keyid rpkg configuration option, commit '
            'and tag will be created signed with the configured '
            'GPG key.', help='Commit changes')
        group = commit_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--message', '-m', default=None,
            help='Use the given <msg> as the commit message summary')
        group.add_argument(
            '--file', '-F', default=None,
            help='Take the commit message from the given file')
        commit_parser.add_argument(
            '--all', '-a', action='store_true', default=False,
            help='Tell the command to automatically stage files '
            'that have been modified or deleted.')
        commit_parser.add_argument(
            '--amend', action='store_true', default=False,
            help='Replace the tip of the current branch by '
            'creating a new commit.')
        commit_parser.add_argument(
            '--with-changelog', action='store_true', default=None,
            help='Get the last changelog from SPEC as commit message '
            'content. This option must be used with -m together.')
        commit_parser.add_argument(
            '--tag', '-t', default=False, action='store_true',
            help='Create a tag for this commit')
        commit_parser.add_argument(
            '--push', '-p', default=False, action='store_true',
            help='Commit and push as one action')
        commit_parser.add_argument(
            'files', nargs='*', default=[], metavar='file',
            help='Optional list of specific files to commit')
        commit_parser.add_argument(
            '--signoff', '-s', default=False, action='store_true',
            help='Include a signed-off-by')
        commit_parser.set_defaults(command=self.commit)

        ci_parser = subparsers.add_parser(
            'ci', parents=[commit_parser], conflict_handler='resolve',
            help='Alias for commit')
        ci_parser.set_defaults(command=self.commit)

    def register_compile(self, subparsers):
        """Register the compile target"""
        compile_parser = subparsers.add_parser(
            'compile', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='This command calls rpmbuild to compile the source. '
            'By default the prep and configure stages will be done as well, '
            'unless the short-circuit option is used.',
            help='Local test rpmbuild compile')
        compile_parser.add_argument(
            '--short-circuit', action='store',
            help='short-circuit compile')
        compile_parser.add_argument(
            '--nocheck', action='store_true',
            help='nocheck compile')
        compile_parser.set_defaults(command=self.compile)

    def register_copr_build(self, subparsers):
        """Register the copr-build target"""
        description="""
        Build package in Copr build system.

        You need to have copr-cli installed and also set up Copr API key.
        For more information about the API key, see API KEY section in
        copr-cli(1) man page.

        NOTE: This method of building the package does not support
        with, without, and rpmdefines rpkg configuration options.
        These options are used to build an srpm which is then
        sent to Copr but rpm build in Copr happens without
        these options already, which is important for the
        resulting build. If you want these options applied,
        you should rather setup auto-rebuilding in Copr
        that will react to new pushes into your repository.
        """
        copr_parser = subparsers.add_parser(
            'copr-build', parents=[self.spec_common_parser],
            formatter_class=argparse.RawDescriptionHelpFormatter,
            description=description,
            help='Build package in Copr build system')
        copr_parser.add_argument(
            '--config', required=False,
            metavar='CONFIG', dest='copr_config',
            help="Path to an alternative Copr configuration file")
        copr_parser.add_argument(
            '--wait',  '-w', action='store_true', default=False,
            help="Wait for build to finish.")
        copr_parser.add_argument(
            'project', nargs='?',
            help='Name of the project in format OWNER/PROJECT. You can '
            'omit it if you have specified copr_project configuration '
            'option.')
        copr_parser.set_defaults(command=self.copr_build)

        build_parser = subparsers.add_parser(
            'build', parents=[copr_parser], conflict_handler='resolve',
            help='Alias for copr-build', description=description)
        build_parser.set_defaults(command=self.copr_build)

    def register_diff(self, subparsers):
        """Register the diff target"""
        diff_parser = subparsers.add_parser(
            'diff', description='Use git diff to show changes that have been '
            'made to tracked files. By default cached changes (changes that '
            'have been git added) will not be shown.',
            help='Show changes between the last commit and working tree, etc.')
        diff_parser.add_argument(
            '--cached', action='store_true',
            help='View staged changes')
        diff_parser.add_argument(
            'files', action='store', nargs='*', default=[], metavar='file',
            help='Optionally diff specific files')
        diff_parser.set_defaults(command=self.diff)

    def register_gimmespec(self, subparsers):
        """Register the gimmespec target"""
        gimmespec_parser = subparsers.add_parser(
            'gimmespec', description='Print auto-located input spec file name',
            help='Print auto-located input spec file name')
        gimmespec_parser.set_defaults(command=self.gimmespec)

    def register_import_srpm(self, subparsers):
        """Register the import target"""
        import_parser = subparsers.add_parser(
            'import', description='This will extract sources, patches, '
            'and the spec file from an srpm and overwrites the current package '
            'content accordingly. It will import to the current branch by '
            'default. Newly discovered sources will be uploaded and '
            '.gitignore and sources file updated accordingly.',
            help='Import srpm content')
        import_parser.add_argument(
            'srpm', action='store',
            help='Source rpm to import. It can be also an URL in which '
            'case the target srpm will be downloaded first.')
        import_parser.add_argument(
            '--repo-path', action='store',
            help='Path to git repository for which to upload '
            'sources. If not specified, it will be automatically '
            'derived from remote URL configured for the current '
            'branch.')
        import_parser.add_argument(
            '--skip-diffs', action='store_true',
            help="Don't show diffs when importing.")
        import_parser.add_argument(
            '--offline', action='store_true',
            help='Do not actually upload the files but do all the rest.')
        import_parser.set_defaults(command=self.import_srpm)

    def register_install(self, subparsers):
        """Register the install target"""
        install_parser = subparsers.add_parser(
            'install', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='This will call rpmbuild to run the install section. '
            'All leading sections will be processed as well, unless the '
            'short-circuit option is used.',
            help='Local test rpmbuild install')
        install_parser.add_argument(
            '--short-circuit', action='store',
            help='short-circuit install')
        install_parser.add_argument(
            '--nocheck', action='store_true',
            help='nocheck install')
        install_parser.set_defaults(command=self.install)

    def register_lint(self, subparsers):
        """Register the lint target"""
        lint_parser = subparsers.add_parser(
            'lint', description='Rpmlint can be configured using '
            'the --rpmlint-config/-r option or by setting rpkg.rpmlint_config '
            'variable.',
            help='Run rpmlint against rendered spec '
            'and build products if present. Needs rpmlint to be installed.')
        lint_parser.add_argument(
            '--info', '-i', default=False, action='store_true',
            help='Display explanations for reported messages')
        lint_parser.add_argument(
            '--rpmlint-config', '-r', default=None,
            help='Use a specific configuration file for rpmlint')
        lint_parser.add_argument(
            'target_path', default=None, nargs='?',
            help='Directory with rpm items to be linted or just '
            'a single rpm item like a .spec file. if not specified, '
            'the latest package directory from base output path will '
            'be selected.')
        lint_parser.set_defaults(command=self.lint)

    def register_local(self, subparsers):
        """Register the local target"""
        local_parser = subparsers.add_parser(
            'local', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Locally test run of rpmbuild producing binary RPMs.',
            help='Local test rpmbuild binary')
        local_parser.add_argument(
            '--nocheck', action='store_true',
            help='nocheck local build')
        local_parser.set_defaults(command=self.local)

    def register_log(self, subparsers):
        """Register the log target"""
        log_parser = subparsers.add_parser(
            'log', description='Display git log history. '
            'Can be configured by git.log_graph and '
            'git.log_all rpkg configuration options.',
            help='Display git log history')
        log_parser.add_argument(
            '--graph', action='store_true',
            help='Draw a text-based graphical representation '
            'of the commit history. By default enabled by '
            'rpkg configuration option git.log_graph.')
        log_parser.add_argument(
            '--all', action='store_true',
            help='Pretend as if all the refs in refs/, along with HEAD, '
            'are listed on the command line as a revision. By default '
            'enabled by rpkg configuration option git.log_all.')
        log_parser.add_argument(
            '--show-signature', action='store_true',
            help='Show gpg signatures and check their '
            'validity.')
        log_parser.add_argument(
            'positionals', nargs='*', metavar='revision',
            help='Specific revisions or revision range to be shown.')
        log_parser.add_argument(
            # Dummy argument. Parsing is done by the previous argument.
            'positionals', nargs='*', action=DoNothing, metavar='path',
            help='Limit commits to those that touch the specific path(s). '
            'You might need to use a -- to explicitly separate paths '
            'from revisions. In that case, you also need to insert another -- '
            'to first separate options from all the positional arguments.')
        log_parser.set_defaults(command=self.log)

    def register_merge(self, subparsers):
        """Register the merge target"""
        merge_parser = subparsers.add_parser(
            'merge', description='Incorporates changes from the named '
            'commits (since the time their histories diverged '
            'from the current branch) into the current branch.',
            help='Incorporates changes from the named commits '
            'into the current branch')
        merge_parser.add_argument(
            '--no-commit', action='store_true',
            help='Perform the merge but do not auto-commit.')
        merge_parser.add_argument(
            '--squash', action='store_true',
            help='Produce the working tree and index state as if a real '
            'merge happened but do not auto-commit. The next commit will '
            'have only a single parent and contain all the merged changes.')
        merge_parser.add_argument(
            '--message', '-m', default=None,
            help='Use the given <msg> as the commit message summary.')
        merge_parser.add_argument(
            '--file', '-F', default=None,
            help='Take the commit message from the given file.')
        merge_parser.add_argument(
            '--signoff', action='store_true',
            help='Add Signed-off-by line by the committer.')
        merge_parser.add_argument(
            'commits', nargs='*', metavar='commit', default=[],
            help='Commits, usually other branch heads, to merge into '
            'our branch. If no commit is given, merge the remote-tracking '
            'branches that the current branch is configured to use as its '
            'upstream.')
        merge_parser.add_argument(
            '--abort', action='store_true',
            help='Abort the current conflict resolution process, '
            'and try to reconstruct the pre-merge state.')
        merge_parser.add_argument(
            '--continue', action='store_true', dest='cont',
            help='When all your merge conflicts have been resolved, '
            'you can conclude the merge by running this command.')
        merge_parser.set_defaults(command=self.merge)

    def register_new(self, subparsers):
        """Register the new target"""
        new_parser = subparsers.add_parser(
            'new', parents=[self.spec_common_parser],
            description='This will use git to show a diff '
            'of all the changes (even uncommitted changes) '
            'since the last package tag was created.',
            help='Diff against last package tag')
        new_parser.set_defaults(command=self.new)

    def register_nvr(self, subparsers):
        """Register the nvr target"""
        nvr_parser = subparsers.add_parser(
            'nvr', parents=[self.spec_common_parser],
            description='Print package name-version-release',
            help='Print package name-version-release')
        nvr_parser.set_defaults(command=self.nvr)

    def register_prep(self, subparsers):
        """Register the prep target"""
        prep_parser = subparsers.add_parser(
            'prep', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Use rpmbuild to "prep" the sources (unpack the source '
            'archive(s) and apply any patches).',
            help='Prep test rpmbuild binary')
        prep_parser.set_defaults(command=self.prep)

    def register_pull(self, subparsers):
        """Register the pull target"""
        pull_parser = subparsers.add_parser(
            'pull', description='This command uses git to fetch remote '
            'changes and apply them to the current working copy. A rebase '
            'option is available which can be used to avoid merges.',
            help='Pull changes from the remote repository '
            'and update the working copy')
        pull_parser.add_argument(
            '--rebase', action='store_true',
            help='Rebase the locally committed changes on top of the remote '
            'changes after fetching. This can avoid a merge commit, but does '
            'rewrite local history.')
        pull_parser.add_argument(
            '--no-rebase', action='store_true',
            help='Do not rebase, overriding .git settings to the contrary')
        pull_parser.set_defaults(command=self.pull)

    def register_push(self, subparsers):
        """Register the push target"""
        push_parser = subparsers.add_parser(
            'push', description='Push changes and annotated tags '
            'to remote repository.',
            help='Push changes and annotated tags to remote repository')
        push_parser.add_argument(
            '--follow-tags', action='store_true',
            help='Push all the refs that would be pushed without this '
            'option and also push annotated tags reachable from the '
            'refs that are pushed by default. This option is passed '
            'by default if git.push_follow_tags configuration variable '
            'is set to true.')
        push_parser.add_argument(
            '--force', '-f', action='store_true',
            help='Force push')
        push_parser.set_defaults(command=self.push)

    def register_reset(self, subparsers):
        """Register the reset target"""
        reset_parser = subparsers.add_parser(
            'reset', description='Reset HEAD to the specified commit '
            'and update index and/or working tree. Alternatively, if '
            'path is passed instead of commit, reset the respective '
            'index entry to its state at HEAD.',
            help='Reset branch tip, index, or the working tree')
        group = reset_parser.add_mutually_exclusive_group()
        group.add_argument(
            '--soft', action='store_true',
            help='Does not touch the index file or the working tree. '
            'Only resets the head to the specified commit.')
        group.add_argument(
            '--mixed', action='store_true',
            help='Resets the index but not the working tree. '
            'This is the default action.')
        group.add_argument(
            '--hard', action='store_true',
            help=' Resets the index and working tree. Any changes '
            'to tracked files in the working tree since the specified '
            'commit are discarded.')
        reset_parser.add_argument(
            'positional', nargs='?', action='store', metavar='commit|path',
            help='Commit or path. If path is passed, the respective '
            'index entry is updated to its HEAD state, otherwise the action '
            'depends on the specified option switch (--mixed is implicit).')
        reset_parser.set_defaults(command=self.reset)

    def register_sources(self, subparsers):
        """Register the sources target"""
        sources_parser = subparsers.add_parser(
            'sources', description='Download source files from '
            'lookaside cache.',
            help='Download source files from lookaside cache')
        sources_parser.add_argument(
            '--repo-path', action='store',
            help='Path to git repository for which to download '
            'sources. If not specified, it will be automatically '
            'derived from remote URL configured for the current '
            'branch.')
        sources_parser.add_argument(
            '--outdir', action='store',
            help='Directory to put downloaded files into '
            '(defaults to cwd)')
        sources_parser.set_defaults(command=self.sources)

    def register_spec(self, subparsers):
        """Register the spec target"""
        spec_parser = subparsers.add_parser(
            'spec', parents=[self.spec_common_parser],
            description='Preprocess a given spec template '
            'and put the resulting spec file into outdir. The spec '
            'template is simply copied into the outdir if '
            'spec preprocessing is turned off or if the input spec '
            'template does not contain any preprocessor macros.',
            help='Generate spec file from a spec template')
        spec_parser.add_argument(
            '--sources', action='store_true', default=False,
            help='Enable source generation by the input '
            'spec template. By default disabled.')
        spec_parser.add_argument(
            '--print', '-p', action='store_true', default=False,
            help='Print rendered spec file to stdout.')
        spec_parser.set_defaults(command=self.spec)

    def register_srpm(self, subparsers):
        """Register the srpm target"""
        srpm_parser = subparsers.add_parser(
            'srpm', parents=[self.rpm_common_parser, self.spec_common_parser],
            description='Create a source rpm',
            help='Create a source rpm')
        srpm_parser.set_defaults(command=self.srpm)

    def register_status(self, subparsers):
        """Register the status target"""
        status_parser = subparsers.add_parser(
            'status', description='Display git status of the '
            'current working tree and index.',
            help='Display git status of the current working '
            'tree and index')
        status_parser.add_argument(
            'pathspecs', nargs='*', metavar='pathspec', default=[],
            help='Limit status information to the specific subpaths.')
        status_parser.set_defaults(command=self.status)

    def register_switch(self, subparsers):
        """Register the switch target"""
        switch_parser = subparsers.add_parser(
            'switch', description='Switch to a given git branch. '
            'If --detach option is given, also tag, or a commit can be used '
            'as a target for switching. Additionally, this command can be '
            'used to list the existing local and remote branches or to '
            'fetch new refs from remote.',
            help='Work with branches')
        switch_parser.add_argument(
            'target', action='store', nargs='?', default=None,
            help='Target branch name or alternatively, commit or tag to switch to. '
            'If provided with a remote branch name that does not have a local match, '
            'it creates one and sets it to track the corresponding remote branch.')
        switch_parser.add_argument(
            '--create', '-c', action='store_true',
            help='Create a new local branch even if it does not have the '
            'corresponding remote and switch to it.')
        switch_parser.add_argument(
             '--detach', '-d', action='store_true',
            help='Detach HEAD at target commit or tag.')
        switch_parser.add_argument(
            '--list-branches', '-l', action='store_true',
            help='List both remote-tracking branches and local branches. '
            'Implied when target is not specified.')
        switch_parser.add_argument(
            '--fetch', action='store_true',
            help='Fetch branches and tags from remote before switch.')
        switch_parser.set_defaults(command=self.switch)

    def register_tag(self, subparsers):
        """Register the tag target"""
        tag_parser = subparsers.add_parser(
            'tag', description='This command uses git to '
            'create, list, or delete tags. If you have '
            'set git.gpg_keyid in rpkg configuraion, the tags '
            'will be created signed.', help='Manage git tags')
        tag_parser.add_argument(
            '--list', '-l', default=False, action='store_true',
            help='List all tags with a given pattern, or '
            'all if not pattern is given')
        tag_parser.add_argument(
            '--delete', '-d', default=False, action='store_true',
            help='Delete a tag')
        tag_parser.add_argument(
            '--skip-edit', '-s', default=False, action='store_true',
            help='Skip changelog editing when creating a tag.')
        tag_parser.add_argument(
            '--name', '-n', default='', action='store',
            help='Set package name part of the tag name to specific value.')
        tag_parser.add_argument(
            '--version', '-v', default='', action='store',
            help='Set version part of the tag name to specific value.')
        tag_parser.add_argument(
            '--release', '-r', default='', action='store',
            help='Set release part of the tag name to specific value.')
        tag_parser.add_argument(
            'tag', nargs='?', default=None, help='Name of the tag. '
            'Overrides any of the --name, --version, --release args '
            'if specified')
        tag_parser.set_defaults(command=self.tag)

    def register_unused_patches(self, subparsers):
        """Register the unused-patches target"""
        unused_patches_parser = subparsers.add_parser(
            'unused-patches', description='Print list of patches in current '
            'working directory not referenced by name in the specfile.',
            help='Print list of patches in CWD not referenced by name '
            'in the specfile')
        unused_patches_parser.set_defaults(command=self.unused_patches)

    def register_upload(self, subparsers):
        """Register the upload target"""
        upload_parser = subparsers.add_parser(
            'upload', description='This command will upload new source '
            'file(s) to lookaside cache. Source file names are appended '
            'to sources file, and .gitignore will be also updated with '
            'new uploaded file(s). Please remember to commit them.',
            help='Upload source files')
        upload_parser.add_argument(
            '--repo-path', action='store',
            help='Path to git repository for which to upload '
            'sources. If not specified, it will be automatically '
            'derived from remote URL configured for the current '
            'branch.')
        upload_parser.add_argument(
            '--offline', action='store_true',
            help='Do not actually upload the files but do all the rest, '
            'which involves updating .gitignore and sources file.')
        upload_parser.add_argument(
            'files', nargs='+', default=[], metavar='file',
            help='List of files to upload')
        upload_parser.set_defaults(command=self.upload)

    def register_verify_files(self, subparsers):
        """Register the verify-files target"""
        verify_files_parser = subparsers.add_parser(
            'verify-files', description='Locally run \'rpmbuild -bl\' to '
            'verify the spec file\'s %%files sections. This requires a '
            'successful run of the install commaned in advance.',
            help='Locally verify %%files section')
        verify_files_parser.add_argument(
            'target_dir', default=None, nargs='?',
            help='Directory with build results to verify. '
            'If not specified, the latest package directory from '
            'base output path will be selected.')
        verify_files_parser.set_defaults(command=self.verify_files)

    @cached_property
    def rpm_common_parser(self):
        """Create a common parser for rpm commands"""
        rpm_common_parser = argparse.ArgumentParser(
            'rpm_common', add_help=False)
        rpm_common_parser.add_argument(
            '--target', default=None,
            help='Prep for a specific arch or arch-vendor-os.')
        rpm_common_parser.add_argument(
            '--with', metavar='OPTION', action='append', default=[],
            help='Enable %%bcond spec OPTION. '
            'Can be specified multiple times.')
        rpm_common_parser.add_argument(
            '--without', metavar='OPTION', action='append', default=[],
            help='Disable %%bcond spec OPTION. '
            'Can be specified multiple times.')
        return rpm_common_parser

    @cached_property
    def spec_common_parser(self):
        """Create a common parser for commands that need a spec file"""
        spec_common_parser = argparse.ArgumentParser(
            'spec_common', add_help=False)
        spec_common_parser.add_argument(
            '--outdir', action='store', default=None,
            help='Where to put any generated rpm content. Directory must exist.')
        spec_common_parser.add_argument(
            '--spec', action='store', default=None,
            help='An input spec template. If not given, the spec template is '
            'auto-located in <path> as a file ending either with ".spec.rpkg" '
            'or ".spec" suffix (in this order). The directory of the template '
            'then also becomes search location for the associated sources and '
            'rpkg.conf file. By default, these files are being looked up in '
            '<path>. Essentially, this argument changes the subpackage we work '
            'with and which is determined by the --path argument initially '
            '(with CWD as the default value for it).')
        return spec_common_parser
